import logging
from typing import List

from eth_utils import decode_hex, is_same_address
from web3 import Web3
from web3.providers.rpc import RPCProvider

from microraiden.utils import (
    get_private_key,
    get_logs,
    get_event_blocking,
    create_signed_contract_transaction
)

from microraiden.config import CHANNEL_MANAGER_ADDRESS
from microraiden.client.context import Context
from microraiden.client.channel import Channel

log = logging.getLogger(__name__)


class Client:
    def __init__(
            self,
            private_key: str = None,
            key_path: str = None,
            key_password_path: str = None,
            channel_manager_address: str = CHANNEL_MANAGER_ADDRESS,
            web3: Web3 = None
    ) -> None:
        assert private_key or key_path
        assert not private_key or isinstance(private_key, str)

        # Load private key from file if none is specified on command line.
        if not private_key:
            private_key = get_private_key(key_path, key_password_path)
            assert private_key is not None

        self.channels = []  # type: List[Channel]

        # Create web3 context if none is provided, either by using the proxies' context or creating
        # a new one.
        if not web3:
            web3 = Web3(RPCProvider())

        self.context = Context(private_key, web3, channel_manager_address)

        self.sync_channels()

    def __enter__(self):
        return self

    def sync_channels(self):
        """
        Merges locally available channel information, including their current balance signatures,
        with channel information available on the blockchain to make up for local data loss.
        Naturally, balance signatures cannot be recovered from the blockchain.
        """
        filters = {'_sender': self.context.address}
        create = get_logs(
            self.context.channel_manager,
            'ChannelCreated',
            argument_filters=filters
        )
        topup = get_logs(
            self.context.channel_manager,
            'ChannelToppedUp',
            argument_filters=filters
        )
        close = get_logs(
            self.context.channel_manager,
            'ChannelCloseRequested',
            argument_filters=filters
        )
        settle = get_logs(
            self.context.channel_manager,
            'ChannelSettled',
            argument_filters=filters
        )

        channel_key_to_channel = {}

        def get_channel(event) -> Channel:
            sender = event['args']['_sender']
            receiver = event['args']['_receiver']
            block = event['args'].get('_open_block_number', event['blockNumber'])
            assert sender == self.context.address
            return channel_key_to_channel.get((sender, receiver, block), None)

        for c in self.channels:
            channel_key_to_channel[(c.sender, c.receiver, c.block)] = c

        for e in create:
            c = get_channel(e)
            if c:
                c.deposit = e['args']['_deposit']
            else:
                c = Channel(
                    self.context,
                    e['args']['_sender'],
                    e['args']['_receiver'],
                    e['blockNumber'],
                    e['args']['_deposit'],
                    on_settle=lambda channel: self.channels.remove(channel)
                )
                assert c.sender == self.context.address
                channel_key_to_channel[(c.sender, c.receiver, c.block)] = c

        for e in topup:
            c = get_channel(e)
            c.deposit += e['args']['_added_deposit']

        for e in close:
            # Requested closed, not actual closed.
            c = get_channel(e)

            c.update_balance(e['args']['_balance'])
            c.state = Channel.State.settling

        for e in settle:
            c = get_channel(e)
            c.state = Channel.State.closed

        # Forget closed channels.
        self.channels = [
            c for c in channel_key_to_channel.values() if c.state != Channel.State.closed
        ]

        log.debug('Synced a total of {} channels.'.format(len(self.channels)))

    def open_channel(self, receiver_address: str, deposit: int):
        """
        Attempts to open a new channel to the receiver with the given deposit. Blocks until the
        creation transaction is found in a pending block or timeout is reached. The new channel
        state is returned.
        """
        assert isinstance(receiver_address, str)
        assert isinstance(deposit, int)
        assert deposit > 0

        token_balance = self.context.token.call().balanceOf(self.context.address)
        if token_balance < deposit:
            log.error(
                'Insufficient tokens available for the specified deposit ({}/{})'
                .format(token_balance, deposit)
            )
            return None

        current_block = self.context.web3.eth.blockNumber
        log.info('Creating channel to {} with an initial deposit of {} @{}'.format(
            receiver_address, deposit, current_block
        ))

        data = decode_hex(receiver_address)
        tx = create_signed_contract_transaction(
            self.context.private_key,
            self.context.token,
            'transfer',
            [
                self.context.channel_manager.address,
                deposit,
                data
            ]
        )
        self.context.web3.eth.sendRawTransaction(tx)

        log.debug('Waiting for channel creation event on the blockchain...')
        filters = {
            '_sender': self.context.address,
            '_receiver': receiver_address
        }
        event = get_event_blocking(
            self.context.channel_manager,
            'ChannelCreated',
            from_block=current_block + 1,
            argument_filters=filters
        )

        if event:
            log.debug('Event received. Channel created in block {}.'.format(event['blockNumber']))
            channel = Channel(
                self.context,
                event['args']['_sender'],
                event['args']['_receiver'],
                event['blockNumber'],
                event['args']['_deposit'],
                on_settle=lambda c: self.channels.remove(c)
            )
            self.channels.append(channel)
        else:
            log.error('Error: No event received.')
            channel = None

        return channel

    def get_open_channels(self, receiver: str = None) -> List[Channel]:
        """
        Returns all open channels to the given receiver. If no receiver is specified, all open
        channels are returned.
        """
        return [
            c for c in self.channels
            if is_same_address(c.sender, self.context.address) and
            (not receiver or is_same_address(c.receiver, receiver)) and
            c.state == Channel.State.open
        ]

    def get_suitable_channel(
            self, receiver, value, initial_deposit=lambda x: x, topup_deposit=lambda x: x
    ) -> Channel:
        """
        Searches stored channels for one that can sustain the given transfer value. If none is
        found, a possibly open channel is topped up using the topup callable to determine its topup
        value. If both attempts fail, a new channel is created based on the initial deposit
        callable.
        Note: In the realistic case that only one channel is opened per (sender,receiver) pair,
        this method usually performs like this:
        1. Directly return open channel if sufficiently funded.
        2. Topup existing open channel if insufficiently funded.
        3. Create new channel if no open channel exists.
        If topping up or creating fails, this method returns None.
        Channels are topped up just enough so that their remaining capacity equals
        topup_deposit(value).
        """
        open_channels = self.get_open_channels(receiver)
        suitable_channels = [c for c in open_channels if c.is_suitable(value)]

        if suitable_channels:
            # At least one channel with sufficient funds.

            if len(suitable_channels) > 1:
                # This is not impossible but should only happen after bad channel management.
                log.warning(
                    'Warning: {} suitable channels found. '
                    'Choosing the one with the lowest remaining capacity.'
                    .format(len(suitable_channels))
                )

            capacity, channel = min(
                ((c.deposit - c.balance, c) for c in suitable_channels), key=lambda x: x[0]
            )
            log.debug(
                'Found suitable open channel, opened at block #{}.'.format(channel.block)
            )
            return channel

        elif open_channels:
            # Open channel(s) but insufficient funds. Requires topup.

            if len(open_channels) > 1:
                # This is not impossible but should only happen after bad channel management.
                log.warning(
                    'Warning: {} open channels for topup found. '
                    'Choosing the one with the highest remaining capacity.'
                    .format(len(open_channels))
                )

            capacity, channel = max(
                ((c.deposit - c.balance, c) for c in open_channels), key=lambda x: x[0]
            )
            deposit = max(value, topup_deposit(value)) - capacity
            event = channel.topup(deposit)
            return channel if event else None

        else:
            # No open channels to receiver. Create a new one.
            deposit = max(value, initial_deposit(value))
            return self.open_channel(receiver, deposit)



import time

import gevent
import gevent.event
import filelock
import logging
import os
from eth_utils import (
    decode_hex,
    is_same_address
)
from ethereum.exceptions import InsufficientBalance
from web3 import Web3
from web3.contract import Contract

from microraiden.utils import (
    verify_balance_proof,
    privkey_to_addr,
    sign_close,
    create_signed_contract_transaction
)
from microraiden.exceptions import (
    NetworkIdMismatch,
    StateReceiverAddrMismatch,
    StateContractAddrMismatch,
    StateFileLocked,
    NoOpenChannel,
    InsufficientConfirmations,
    InvalidBalanceProof,
    InvalidBalanceAmount,
    InvalidContractVersion,
    NoBalanceProofReceived,
)
from microraiden.config import CHANNEL_MANAGER_CONTRACT_VERSION
from .state import ChannelManagerState
from .blockchain import Blockchain
from .channel import Channel, ChannelState

log = logging.getLogger(__name__)


class ChannelManager(gevent.Greenlet):
    """Manages channels from the receiver's point of view."""

    def __init__(
            self,
            web3: Web3,
            channel_manager_contract: Contract,
            token_contract: Contract,
            private_key: str,
            state_filename: str = None,
            n_confirmations=1
    ) -> None:
        gevent.Greenlet.__init__(self)
        self.blockchain = Blockchain(
            web3,
            channel_manager_contract,
            self,
            n_confirmations=n_confirmations
        )
        self.receiver = privkey_to_addr(private_key)
        self.private_key = private_key
        self.channel_manager_contract = channel_manager_contract
        self.token_contract = token_contract
        self.n_confirmations = n_confirmations
        self.log = logging.getLogger('channel_manager')
        network_id = web3.version.network
        assert privkey_to_addr(self.private_key) == self.receiver.lower()

        # check contract version
        self.check_contract_version()

        if state_filename not in (None, ':memory:') and os.path.isfile(state_filename):
            self.state = ChannelManagerState.load(state_filename)
        else:
            self.state = ChannelManagerState(state_filename)
            self.state.setup_db(
                network_id,
                channel_manager_contract.address,
                self.receiver
            )

        assert self.state is not None
        if state_filename not in (None, ':memory:'):
            self.lock_state = filelock.FileLock(state_filename + '.lock')
            try:
                self.lock_state.acquire(timeout=0)
            except filelock.Timeout:
                raise StateFileLocked("state file %s is locked by another process" %
                                      state_filename)

        if network_id != self.state.network_id:
            raise NetworkIdMismatch("Network id mismatch: state=%d, backend=%d" % (
                                    self.state.network_id, network_id))

        if not is_same_address(self.receiver, self.state.receiver):
            raise StateReceiverAddrMismatch('%s != %s' %
                                            (self.receiver.lower(), self.state.receiver))
        if not is_same_address(self.state.contract_address, channel_manager_contract.address):
            raise StateContractAddrMismatch('%s != %s' % (
                channel_manager_contract.address.lower(), self.state.contract_address.lower()))

        self.log.debug('setting up channel manager, receiver=%s channel_contract=%s' %
                       (self.receiver, channel_manager_contract.address))

    def __del__(self):
        self.stop()

    def _run(self):
        self.blockchain.start()

    def stop(self):
        if self.blockchain.running:
            self.blockchain.stop()
            self.blockchain.join()

    def set_head(self, unconfirmed_head_number, unconfirmed_head_hash,
                 confirmed_head_number, confirmed_head_hash):
        """Set the block number up to which all events have been registered."""
        self.state.update_sync_state(unconfirmed_head_number=unconfirmed_head_number,
                                     unconfirmed_head_hash=unconfirmed_head_hash,
                                     confirmed_head_number=confirmed_head_number,
                                     confirmed_head_hash=confirmed_head_hash)

    # relevant events from the blockchain for receiver from contract

    def event_channel_opened(self, sender, open_block_number, deposit):
        """Notify the channel manager of a new confirmed channel opening."""
        if (sender, open_block_number) in self.channels:
            return  # ignore event if already provessed
        c = Channel(self.state.receiver, sender, deposit, open_block_number)
        c.confirmed = True
        c.state = ChannelState.OPEN
        self.log.info('new channel opened (sender %s, block number %s)', sender, open_block_number)
        self.state.set_channel(c)

    def unconfirmed_event_channel_opened(self, sender, open_block_number, deposit):
        """Notify the channel manager of a new channel opening that has not been confirmed yet."""
        event_already_processed = (sender, open_block_number) in self.unconfirmed_channels
        channel_already_confirmed = (sender, open_block_number) in self.channels
        if event_already_processed or channel_already_confirmed:
            return
        c = Channel(self.state.receiver, sender, deposit, open_block_number)
        c.confirmed = False
        c.state = ChannelState.OPEN
        self.state.set_channel(c)
        self.log.info('unconfirmed channel event received (sender %s, block_number %s)',
                      sender, open_block_number)

    def event_channel_close_requested(self, sender, open_block_number, balance, settle_timeout):
        assert settle_timeout >= 0
        """Notify the channel manager that a the closing of a channel has been requested."""
        if (sender, open_block_number) not in self.channels:
            self.log.warning(
                'attempt to close a non existing channel (sender %ss, block_number %ss)',
                sender,
                open_block_number
            )
            return
        c = self.channels[sender, open_block_number]
        if c.balance > balance:
            self.log.warning('sender tried to cheat, sending challenge '
                             '(sender %s, block number %s)',
                             sender, open_block_number)
            self.close_channel(sender, open_block_number)  # dispute by closing the channel
        else:
            self.log.info('valid channel close request received '
                          '(sender %s, block number %s, timeout %d)',
                          sender, open_block_number, settle_timeout)
            c.settle_timeout = settle_timeout
            c.is_closed = True
            c.confirmed = True
            c.mtime = time.time()
        self.state.set_channel(c)

    def event_channel_settled(self, sender, open_block_number):
        """Notify the channel manager that a channel has been settled."""
        self.log.info('Forgetting settled channel (sender %s, block number %s)',
                      sender, open_block_number)
        self.state.del_channel(sender, open_block_number)

    def unconfirmed_event_channel_topup(
            self, sender, open_block_number, txhash, added_deposit
    ):
        """Notify the channel manager of a topup with not enough confirmations yet."""
        if (sender, open_block_number) not in self.channels:
            assert (sender, open_block_number) in self.unconfirmed_channels
            self.log.info('Ignoring unconfirmed topup of unconfirmed channel '
                          '(sender %s, block number %s, added %s)',
                          sender, open_block_number, added_deposit)
            return
        self.log.info('Registering unconfirmed deposit top up '
                      '(sender %s, block number %s, added %s)',
                      sender, open_block_number, added_deposit)
        c = self.channels[sender, open_block_number]
        c.unconfirmed_topups[txhash] = added_deposit
        self.state.set_channel(c)

    def event_channel_topup(self, sender, open_block_number, txhash, added_deposit):
        """Notify the channel manager that the deposit of a channel has been topped up."""
        self.log.info(
            'Registering deposit top up (sender %s, block number %s, added deposit %s)',
            sender, open_block_number, added_deposit
        )
        assert (sender, open_block_number) in self.channels
        c = self.channels[sender, open_block_number]
        if c.is_closed is True:
            self.log.warning(
                "Topup of an already closed channel (sender=%s open_block=%d)" %
                (sender, open_block_number)
            )
            return None
        c.deposit += added_deposit
        c.unconfirmed_topups.pop(txhash, None)
        c.mtime = time.time()
        self.state.set_channel(c)

    # end events ####

    def close_channel(self, sender, open_block_number):
        """Close and settle a channel."""
        if not (sender, open_block_number) in self.channels:
            self.log.warning(
                "attempt to close a non-registered channel (sender=%s open_block=%s" %
                (sender, open_block_number)
            )
            return
        c = self.channels[sender, open_block_number]
        if c.last_signature is None:
            raise NoBalanceProofReceived('Cannot close a channel without a balance proof.')
        # send closing tx
        raw_tx = create_signed_contract_transaction(
            self.private_key,
            self.channel_manager_contract,
            'uncooperativeClose',
            [
                self.state.receiver,
                open_block_number,
                c.balance,
                decode_hex(c.last_signature)
            ]
        )

        # update local state
        c.is_closed = True
        c.mtime = time.time()
        self.state.set_channel(c)

        try:
            txid = self.blockchain.web3.eth.sendRawTransaction(raw_tx)
            self.log.info('sent channel close(sender %s, block number %s, tx %s)',
                          sender, open_block_number, txid)
        except InsufficientBalance:
            c.state = ChannelState.CLOSE_PENDING
            self.state.set_channel(c)
            raise

    def force_close_channel(self, sender, open_block_number):
        """Forcibly remove a channel from our channel state"""
        try:
            self.close_channel(sender, open_block_number)
            return
        except NoBalanceProofReceived:
            c = self.channels[sender, open_block_number]
            c.is_closed = True
            self.state.set_channel(c)

    def sign_close(self, sender, open_block_number, balance):
        """Sign an agreement for a channel closing."""
        if (sender, open_block_number) not in self.channels:
            raise NoOpenChannel('Channel does not exist or has been closed'
                                '(sender=%s, open_block_number=%d)' % (sender, open_block_number))
        c = self.channels[sender, open_block_number]
        if c.is_closed:
            raise NoOpenChannel('Channel closing has been requested already.')
        assert balance is not None
        if c.last_signature is None:
            raise NoBalanceProofReceived('Payment has not been registered.')
        if balance != c.balance:
            raise InvalidBalanceProof('Requested closing balance does not match latest one.')
        c.is_closed = True
        c.mtime = time.time()
        receiver_sig = sign_close(self.private_key, c.last_signature)
        self.state.set_channel(c)
        self.log.info('signed cooperative closing message (sender %s, block number %s)',
                      sender, open_block_number)
        return receiver_sig

    def get_locked_balance(self):
        """Get the balance in all channels combined."""
        return sum([c.balance for c in self.channels.values()])

    def get_liquid_balance(self):
        """Get the balance of the receiver in the token contract (not locked in channels)."""
        balance = self.token_contract.call().balanceOf(self.receiver)
        return balance

    def get_eth_balance(self):
        """Get eth balance of the receiver"""
        return self.channel_manager_contract.web3.eth.getBalance(self.receiver)

    def verify_balance_proof(self, sender, open_block_number, balance, signature):
        """Verify that a balance proof is valid and return the sender.

        Does not check the balance itself.

        :returns: the channel
        """
        if (sender, open_block_number) in self.unconfirmed_channels:
            raise InsufficientConfirmations(
                'Insufficient confirmations for the channel '
                '(sender=%s, open_block_number=%d)' % (sender, open_block_number))
        try:
            c = self.channels[sender, open_block_number]
        except KeyError:
            raise NoOpenChannel('Channel does not exist or has been closed'
                                '(sender=%s, open_block_number=%s)' % (sender, open_block_number))
        if c.is_closed:
            raise NoOpenChannel('Channel closing has been requested already.')

        if not is_same_address(
                verify_balance_proof(
                    self.receiver,
                    open_block_number,
                    balance,
                    decode_hex(signature),
                    self.channel_manager_contract.address
                ),
                sender
        ):
            raise InvalidBalanceProof('Recovered signer does not match the sender')
        return c

    def register_payment(self, sender, open_block_number, balance, signature):
        """Register a payment."""
        c = self.verify_balance_proof(sender, open_block_number, balance, signature)
        if balance <= c.balance:
            raise InvalidBalanceAmount('The balance must not decrease.')
        if balance > c.deposit:
            raise InvalidBalanceProof('Balance must not be greater than deposit')
        received = balance - c.balance
        c.balance = balance
        c.last_signature = signature
        c.mtime = time.time()
        self.state.set_channel(c)
        self.log.debug('registered payment (sender %s, block number %s, new balance %s)',
                       c.sender, open_block_number, balance)
        return c.sender, received

    def reset_unconfirmed(self):
        """Forget all unconfirmed channels and topups to allow for a clean resync."""
        self.state.del_unconfirmed_channels()
        for channel in self.channels.values():
            channel.unconfirmed_topups.clear()
            self.state.set_channel(channel)
        self.state.unconfirmed_head_number = self.state.confirmed_head_number
        self.state.unconfirmed_head_hash = self.state.confirmed_head_hash

    @property
    def channels(self):
        return self.state.channels

    @property
    def unconfirmed_channels(self):
        return self.state.unconfirmed_channels

    @property
    def pending_channels(self):
        return self.state.pending_channels

    def channels_to_dict(self):
        """Export all channels as a dictionary."""
        d = {}
        for sender, block_number in self.channels:
            channel = self.channels[sender, block_number]
            channel_dict = {
                'deposit': channel.deposit,
                'balance': channel.balance,
                'mtime': channel.mtime,
                'ctime': channel.ctime,
                'settle_timeout': channel.settle_timeout,
                'last_signature': channel.last_signature,
                'is_closed': channel.is_closed
            }
            if sender not in d:
                d[sender] = {}
            d[sender][block_number] = channel_dict
        return d

    def unconfirmed_channels_to_dict(self):
        """Export all unconfirmed channels as a dictionary."""
        d = {}
        for sender, block_number in self.unconfirmed_channels:
            channel = self.unconfirmed_channels[sender, block_number]
            channel_dict = {
                'deposit': channel.deposit,
                'ctime': channel.ctime
            }
            if sender not in d:
                d[sender] = {}
            d[sender][block_number] = channel_dict
        return d

    def wait_sync(self):
        self.blockchain.wait_sync()

    def node_online(self):
        return self.blockchain.is_connected.is_set()

    def get_token_address(self):
        return self.token_contract.address

    def check_contract_version(self):
        deployed_contract_version = self.channel_manager_contract.call().version()
        if deployed_contract_version != CHANNEL_MANAGER_CONTRACT_VERSION:
            raise InvalidContractVersion("Incompatible contract version: expected=%s deployed=%s" %
                                         (CHANNEL_MANAGER_CONTRACT_VERSION,
                                          deployed_contract_version))

    def close_pending_channels(self):
        """Close all channels that are in CLOSE_PENDING state.
        This state happens if the receiver's eth balance is not enough to
            close channel on-chain."""
        for sender, open_block_number in self.pending_channels.keys():
            self.close_channel(sender, open_block_number)  # dispute by closing the channel



import logging
import time
from typing import Callable

from eth_utils import is_same_address, decode_hex
from requests import Response

from microraiden.header import HTTPHeaders
from microraiden.client import Client
from microraiden.utils import verify_balance_proof
from .http_client import HTTPClient

log = logging.getLogger(__name__)


class DefaultHTTPClient(HTTPClient):
    def __init__(
            self,
            client: Client,
            retry_interval: float = 5,
            initial_deposit: Callable[[int], int] = lambda price: 10 * price,
            topup_deposit: Callable[[int], int] = lambda price: 5 * price
    ) -> None:
        HTTPClient.__init__(self, client)
        self.retry_interval = retry_interval
        self.initial_deposit = initial_deposit
        self.topup_deposit = topup_deposit

    def on_insufficient_confirmations(
            self,
            method: str,
            url: str,
            response: Response,
            **kwargs
    ) -> bool:
        log.warning(
            'Newly created channel does not have enough confirmations yet. Retrying in {} seconds.'
            .format(self.retry_interval)
        )
        time.sleep(self.retry_interval)
        return True

    def on_invalid_amount(
            self,
            method: str,
            url: str,
            response: Response,
            **kwargs
    ) -> bool:
        log.debug('Server claims an invalid amount sent.')
        balance_sig = response.headers.get(HTTPHeaders.BALANCE_SIGNATURE)
        if balance_sig:
            balance_sig = decode_hex(balance_sig)
        last_balance = int(response.headers.get(HTTPHeaders.SENDER_BALANCE))

        channel = self.get_channel(url)
        verified = balance_sig and is_same_address(
            verify_balance_proof(
                channel.receiver,
                channel.block,
                last_balance,
                balance_sig,
                self.client.context.channel_manager.address
            ),
            channel.sender
        )

        if verified:
            if last_balance == channel.balance:
                log.error(
                    'Server tried to disguise the last unconfirmed payment as a confirmed payment.'
                )
                return False
            else:
                log.debug(
                    'Server provided proof for a different channel balance ({}). Adopting.'.format(
                        last_balance
                    )
                )
                channel.update_balance(last_balance)
        else:
            log.debug(
                'Server did not provide proof for a different channel balance. Reverting to 0.'
            )
            channel.update_balance(0)

        return self.on_payment_requested(method, url, response, **kwargs)

    def on_payment_requested(
            self,
            method: str,
            url: str,
            response: Response,
            **kwargs
    ) -> bool:
        receiver = response.headers[HTTPHeaders.RECEIVER_ADDRESS]
        price = int(response.headers[HTTPHeaders.PRICE])
        assert price > 0

        log.debug('Preparing payment of price {} to {}.'.format(price, receiver))

        channel = self.get_channel(url)
        if not channel or not channel.is_suitable(price):
            new_channel = self.client.get_suitable_channel(
                receiver, price, self.initial_deposit, self.topup_deposit
            )

            if channel and new_channel != channel:
                # This should only happen if there are multiple open channels to the target.
                log.warning(
                    'Channels switched. Previous balance proofs not applicable to new channel.'
                )

            channel = new_channel
            self.set_channel(url, channel)

        if not channel:
            log.error("No channel could be created or sufficiently topped up.")
            return False

        channel.create_transfer(price)
        log.debug(
            'Sending new balance proof. New channel balance: {}/{}'
            .format(channel.balance, channel.deposit)
        )

        return True

    def on_http_error(self, method: str, url: str, response: Response, **kwargs):
        log.warning(
            'Unexpected server error, status code {}. Retrying in {} seconds.'.format(
                response.status_code, self.retry_interval
            )
        )
        time.sleep(self.retry_interval)
        return True



import requests
import sys
import gevent
import logging

from ethereum.exceptions import InsufficientBalance
from web3 import Web3
from web3.contract import Contract
from web3.exceptions import BadFunctionCallOutput

import microraiden.config as config
from microraiden.utils import get_logs


class Blockchain(gevent.Greenlet):
    """Class that watches the blockchain and relays events to the channel manager."""
    poll_interval = 2

    def __init__(
            self,
            web3: Web3,
            channel_manager_contract: Contract,
            channel_manager,
            n_confirmations,
            sync_chunk_size=100 * 1000
    ):
        gevent.Greenlet.__init__(self)
        self.web3 = web3
        self.channel_manager_contract = channel_manager_contract
        self.cm = channel_manager
        self.n_confirmations = n_confirmations
        self.log = logging.getLogger('blockchain')
        self.wait_sync_event = gevent.event.Event()
        self.is_connected = gevent.event.Event()
        self.sync_chunk_size = sync_chunk_size
        self.running = False
        #  insufficient_balance
        #  - set to true if for some reason tx can't be send
        #  - set to false when all pending closes are settled
        #  - used to dispute/close channels that are in CHALLENGED state after manager
        #     has ether to spend again
        self.insufficient_balance = False

    def _run(self):
        self.running = True
        self.log.info('starting blockchain polling (interval %ss)', self.poll_interval)
        while self.running:
            if self.insufficient_balance:
                self.insufficient_balance_recover()
            try:
                self._update()
                self.is_connected.set()
                if self.wait_sync_event.is_set():
                    gevent.sleep(self.poll_interval)
            except requests.exceptions.ConnectionError as e:
                endpoint = self.web3.currentProvider.endpoint_uri
                self.log.warning(
                    'Ethereum node (%s) refused connection. Retrying in %d seconds.' %
                    (endpoint, self.poll_interval)
                )
                gevent.sleep(self.poll_interval)
                self.is_connected.clear()
        self.log.info('stopped blockchain polling')

    def stop(self):
        self.running = False

    def wait_sync(self):
        self.wait_sync_event.wait()

    def _update(self):
        current_block = self.web3.eth.blockNumber
        # reset unconfirmed channels in case of reorg
        if self.wait_sync_event.is_set():  # but not on first sync
            if current_block < self.cm.state.unconfirmed_head_number:
                self.log.info('chain reorganization detected. '
                              'Resyncing unconfirmed events (unconfirmed_head=%d) [@%d]' %
                              (self.cm.state.unconfirmed_head_number, self.web3.eth.blockNumber))
                self.cm.reset_unconfirmed()
            try:
                # raises if hash doesn't exist (i.e. block has been replaced)
                self.web3.eth.getBlock(self.cm.state.unconfirmed_head_hash)
            except ValueError:
                self.log.info('chain reorganization detected. '
                              'Resyncing unconfirmed events (unconfirmed_head=%d) [@%d]. '
                              '(getBlock() raised ValueError)' %
                              (self.cm.state.unconfirmed_head_number, current_block))
                self.cm.reset_unconfirmed()

            # in case of reorg longer than confirmation number fail
            try:
                self.web3.eth.getBlock(self.cm.state.confirmed_head_hash)
            except ValueError:
                self.log.critical('events considered confirmed have been reorganized')
                assert False  # unreachable as long as confirmation level is set high enough

        if self.cm.state.confirmed_head_number is None:
            self.cm.state.update_sync_state(confirmed_head_number=-1)
        if self.cm.state.unconfirmed_head_number is None:
            self.cm.state.update_sync_state(unconfirmed_head_number=-1)
        new_unconfirmed_head_number = self.cm.state.unconfirmed_head_number + self.sync_chunk_size
        new_unconfirmed_head_number = min(new_unconfirmed_head_number, current_block)
        new_confirmed_head_number = max(new_unconfirmed_head_number - self.n_confirmations, 0)

        # return if blocks have already been processed
        if (self.cm.state.confirmed_head_number >= new_confirmed_head_number and
                self.cm.state.unconfirmed_head_number >= new_unconfirmed_head_number):
            return

        # filter for events after block_number
        filters_confirmed = {
            'from_block': self.cm.state.confirmed_head_number + 1,
            'to_block': new_confirmed_head_number,
            'argument_filters': {
                '_receiver': self.cm.state.receiver
            }
        }
        filters_unconfirmed = {
            'from_block': self.cm.state.unconfirmed_head_number + 1,
            'to_block': new_unconfirmed_head_number,
            'argument_filters': {
                '_receiver': self.cm.state.receiver
            }
        }
        self.log.debug(
            'filtering for events u:%s-%s c:%s-%s @%d',
            filters_unconfirmed['from_block'],
            filters_unconfirmed['to_block'],
            filters_confirmed['from_block'],
            filters_confirmed['to_block'],
            current_block
        )

        # unconfirmed channel created
        logs = get_logs(
            self.channel_manager_contract,
            'ChannelCreated',
            **filters_unconfirmed
        )
        for log in logs:
            assert log['args']['_receiver'] == self.cm.state.receiver
            sender = log['args']['_sender']
            deposit = log['args']['_deposit']
            open_block_number = log['blockNumber']
            self.log.debug(
                'received unconfirmed ChannelCreated event (sender %s, block number %s)',
                sender,
                open_block_number
            )
            self.cm.unconfirmed_event_channel_opened(sender, open_block_number, deposit)

        # channel created
        logs = get_logs(
            self.channel_manager_contract,
            'ChannelCreated',
            **filters_confirmed
        )
        for log in logs:
            assert log['args']['_receiver'] == self.cm.state.receiver
            sender = log['args']['_sender']
            deposit = log['args']['_deposit']
            open_block_number = log['blockNumber']
            self.log.debug('received ChannelOpened event (sender %s, block number %s)',
                           sender, open_block_number)
            self.cm.event_channel_opened(sender, open_block_number, deposit)

        # unconfirmed channel top ups
        logs = get_logs(
            self.channel_manager_contract,
            'ChannelToppedUp',
            **filters_unconfirmed
        )
        for log in logs:
            assert log['args']['_receiver'] == self.cm.state.receiver
            txhash = log['transactionHash']
            sender = log['args']['_sender']
            open_block_number = log['args']['_open_block_number']
            added_deposit = log['args']['_added_deposit']
            self.log.debug(
                'received top up event (sender %s, block number %s, deposit %s)',
                sender,
                open_block_number,
                added_deposit
            )
            self.cm.unconfirmed_event_channel_topup(
                sender,
                open_block_number,
                txhash,
                added_deposit
            )

        # confirmed channel top ups
        logs = get_logs(
            self.channel_manager_contract,
            'ChannelToppedUp',
            **filters_confirmed
        )
        for log in logs:
            assert log['args']['_receiver'] == self.cm.state.receiver
            txhash = log['transactionHash']
            sender = log['args']['_sender']
            open_block_number = log['args']['_open_block_number']
            added_deposit = log['args']['_added_deposit']
            self.log.debug(
                'received top up event (sender %s, block number %s, added deposit %s)',
                sender,
                open_block_number,
                added_deposit
            )
            self.cm.event_channel_topup(sender, open_block_number, txhash, added_deposit)

        # channel settled event
        logs = get_logs(
            self.channel_manager_contract,
            'ChannelSettled',
            **filters_confirmed
        )
        for log in logs:
            assert log['args']['_receiver'] == self.cm.state.receiver
            sender = log['args']['_sender']
            open_block_number = log['args']['_open_block_number']
            self.log.debug('received ChannelSettled event (sender %s, block number %s)',
                           sender, open_block_number)
            self.cm.event_channel_settled(sender, open_block_number)

        # channel close requested
        logs = get_logs(
            self.channel_manager_contract,
            'ChannelCloseRequested',
            **filters_confirmed
        )
        for log in logs:
            assert log['args']['_receiver'] == self.cm.state.receiver
            sender = log['args']['_sender']
            open_block_number = log['args']['_open_block_number']
            if (sender, open_block_number) not in self.cm.channels:
                continue
            balance = log['args']['_balance']
            try:
                timeout = self.channel_manager_contract.call().getChannelInfo(
                    sender,
                    self.cm.state.receiver,
                    open_block_number
                )[2]
            except BadFunctionCallOutput:
                self.log.warning(
                    'received ChannelCloseRequested event for a channel that doesn\'t '
                    'exist or has been closed already (sender=%s open_block_number=%d)'
                    % (sender, open_block_number))
                self.cm.force_close_channel(sender, open_block_number)
                continue
            self.log.debug('received ChannelCloseRequested event (sender %s, block number %s)',
                           sender, open_block_number)
            try:
                self.cm.event_channel_close_requested(sender, open_block_number, balance, timeout)
            except InsufficientBalance:
                self.log.fatal('Insufficient ETH balance of the receiver. '
                               "Can't close the channel. "
                               'Will retry once the balance is sufficient')
                self.insufficient_balance = True
                # TODO: recover

        # update head hash and number
        try:
            new_unconfirmed_head_hash = self.web3.eth.getBlock(new_unconfirmed_head_number).hash
            new_confirmed_head_hash = self.web3.eth.getBlock(new_confirmed_head_number).hash
        except AttributeError:
            self.log.critical("RPC endpoint didn't return proper info for an existing block "
                              "(%d,%d)" % (new_unconfirmed_head_number, new_confirmed_head_number))
            self.log.critical("It is possible that the blockchain isn't fully synced. "
                              "This often happens when Parity is run with --fast or --warp sync.")
            self.log.critical("Can't continue - check status of the ethereum node.")
            sys.exit(1)
        self.cm.set_head(
            new_unconfirmed_head_number,
            new_unconfirmed_head_hash,
            new_confirmed_head_number,
            new_confirmed_head_hash
        )
        if not self.wait_sync_event.is_set() and new_unconfirmed_head_number == current_block:
            self.wait_sync_event.set()

    def insufficient_balance_recover(self):
        balance = self.web3.eth.getBalance(self.cm.receiver)
        if balance < config.PROXY_BALANCE_LIMIT:
            return
        try:
            self.cm.close_pending_channels()
            self.insufficient_balance = False
        except InsufficientBalance:
            self.log.fatal('Insufficient balance when trying to'
                           'close pending channels. (balance=%d)'
                           % balance)



from microraiden import DefaultHTTPClient
from microraiden.test.utils.client import patch_on_http_response

from microraiden.test.config import (
    FAUCET_ADDRESS,
)


#  @pytest.mark.skip(reason="waiting for client to return status code")
def test_cheating_client(
        doggo_proxy,
        web3,
        default_http_client: DefaultHTTPClient,
        wait_for_blocks,
        http_doggo_url: str,
):
    patch_on_http_response(default_http_client, abort_on=[502])
    balance = web3.eth.getBalance(doggo_proxy.channel_manager.receiver)
    assert balance > 0
    # remove all receiver's eth
    web3.eth.sendTransaction({'from': doggo_proxy.channel_manager.receiver,
                              'to': FAUCET_ADDRESS,
                              'value': balance - 90000})
    wait_for_blocks(1)
    default_http_client.get(http_doggo_url)
    # proxy is expected to return 502 - it has no funds
    assert default_http_client.last_response.status_code == 502
    web3.eth.sendTransaction({'from': FAUCET_ADDRESS,
                              'to': doggo_proxy.channel_manager.receiver,
                              'value': balance})
    wait_for_blocks(1)
    default_http_client.get(http_doggo_url)
    # now it should proceed normally
    assert default_http_client.last_response.status_code == 200



from microraiden.examples.eth_ticker import ETHTickerProxy
from microraiden.test.config import TEST_RECEIVER_PRIVKEY
import logging

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    proxy = ETHTickerProxy(TEST_RECEIVER_PRIVKEY)
    proxy.app.join()
    proxy.stop()



import getpass
import json
import logging

import os
import stat

from eth_utils import is_hex, decode_hex, encode_hex
from ethereum import keys

log = logging.getLogger(__name__)


def check_permission_safety(path):
    """Check if the file at the given path is safe to use as a state file.

    This checks that group and others have no permissions on the file and that the current user is
    the owner.
    """
    f_stats = os.stat(path)
    return (f_stats.st_mode & (stat.S_IRWXG | stat.S_IRWXO)) == 0 and f_stats.st_uid == os.getuid()


def get_private_key(key_path, password_path=None):
    """Open a JSON-encoded private key and return it

    If a password file is provided, uses it to decrypt the key. If not, the
    password is asked interactively. Raw hex-encoded private keys are supported,
    but deprecated."""

    assert key_path, key_path
    if not os.path.exists(key_path):
        log.fatal("%s: no such file", key_path)
        return None

    if not check_permission_safety(key_path):
        log.fatal("Private key file %s must be readable only by its owner.", key_path)
        return None

    if password_path and not check_permission_safety(password_path):
        log.fatal("Password file %s must be readable only by its owner.", password_path)
        return None

    with open(key_path) as keyfile:
        private_key = keyfile.readline().strip()

        if is_hex(private_key) and len(decode_hex(private_key)) == 32:
            log.warning("Private key in raw format. Consider switching to JSON-encoded")
        else:
            keyfile.seek(0)
            try:
                json_data = json.load(keyfile)
                if password_path:
                    with open(password_path) as password_file:
                        password = password_file.readline().strip()
                else:
                    password = getpass.getpass("Enter the private key password: ")
                private_key = encode_hex(keys.decode_keystore_json(json_data, password))
            except ValueError:
                log.fatal("Invalid private key format or password!")
                return None

    return private_key



import json
import os
from eth_utils import denoms

API_PATH = "/api/1"
GAS_LIMIT = 250000
GAS_PRICE = 20 * denoms.gwei

NETWORK_NAMES = {
    1: 'mainnet',
    2: 'morden',
    3: 'ropsten',
    4: 'rinkeby',
    30: 'rootstock-main',
    31: 'rootstock-test',
    42: 'kovan',
    61: 'etc-main',
    62: 'etc-test',
    1337: 'geth'
}

# address of the default channel manager contract. You can change this using commandline
# option --channel-manager-address when running the proxy
CHANNEL_MANAGER_ADDRESS = '0xeb8f0c65921c6f2b6025d9dbdd9466419cbe56ec'
# absolute path to this directory. Used to find path to the webUI sources
MICRORAIDEN_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
# webUI sources
HTML_DIR = os.path.join(MICRORAIDEN_DIR, 'microraiden', 'webui')
# javascript library for microraiden
JSLIB_DIR = os.path.join(HTML_DIR, 'js')
# url prefix for jslib dir
JSPREFIX_URL = '/js'
# decimals of the token. Any price that's set for the proxy resources is multiplied by this.
TKN_DECIMALS = 10**18  # token decimals

# ethereum node RPC interface should be available here
WEB3_PROVIDER_DEFAULT = "http://127.0.0.1:8545"

# name of the channel manager contract
CHANNEL_MANAGER_ABI_NAME = 'RaidenMicroTransferChannels'
# name of the token contract
TOKEN_ABI_NAME = 'CustomToken'
# compiled contracts path
CONTRACTS_ABI_JSON = 'data/contracts.json'

with open(os.path.join(MICRORAIDEN_DIR, 'microraiden', CONTRACTS_ABI_JSON)) as metadata_file:
    CONTRACT_METADATA = json.load(metadata_file)

# required version of the deployed contract at CHANNEL_MANAGER_ADDRESS.
# Proxy will refuse to start if the versions do not match.
CHANNEL_MANAGER_CONTRACT_VERSION = "1.0.0"
#  proxy will stop serving requests if receiver balance is below PROXY_BALANCE_LIMIT
PROXY_BALANCE_LIMIT = 10**6


# sanity checks
assert PROXY_BALANCE_LIMIT > 0
assert isinstance(PROXY_BALANCE_LIMIT, int)



import json
import types

import pytest
import requests_mock
from eth_utils import encode_hex
from munch import Munch
from requests import Response
from requests.exceptions import SSLError

from microraiden import HTTPHeaders
from microraiden import DefaultHTTPClient
from microraiden.test.utils.client import patch_on_http_response
from microraiden.test.utils.disable_ssl_check import disable_ssl_check


def check_response(response: Response):
    assert response and response.text == 'HI I AM A DOGGO'


def test_full_cycle_success(
        default_http_client: DefaultHTTPClient,
        api_endpoint_address: str,
        token_address: str,
        channel_manager_address: str,
        receiver_address: str
):
    default_http_client.initial_deposit = lambda x: x

    with requests_mock.mock() as server_mock:
        headers1 = Munch()
        headers1.token_address = token_address
        headers1.contract_address = channel_manager_address
        headers1.receiver_address = receiver_address
        headers1.price = '7'

        headers2 = Munch()
        headers2.cost = '7'

        headers1 = HTTPHeaders.serialize(headers1)
        headers2 = HTTPHeaders.serialize(headers2)

        url = 'http://{}/something'.format(api_endpoint_address)
        server_mock.get(url, [
            {'status_code': 402, 'headers': headers1},
            {'status_code': 200, 'headers': headers2, 'text': 'success'}
        ])
        response = default_http_client.get(url)

    # First cycle, request price.
    request = server_mock.request_history[0]
    assert request.path == '/something'
    assert request.method == 'GET'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address

    # Second cycle, pay price.
    request = server_mock.request_history[1]
    assert request.path == '/something'
    assert request.method == 'GET'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address
    assert request.headers['RDN-Balance'] == '7'

    channel = default_http_client.get_channel(url)
    assert channel.balance == 7
    balance_sig_hex = encode_hex(channel.balance_sig)
    assert request.headers['RDN-Balance-Signature'] == balance_sig_hex
    assert channel.balance_sig
    assert response.text == 'success'


def test_full_cycle_adapt_balance(
        default_http_client: DefaultHTTPClient,
        api_endpoint_address: str,
        token_address: str,
        channel_manager_address: str,
        receiver_address: str
):
    # Simulate a lost balance signature.
    client = default_http_client.client
    channel = client.get_suitable_channel(receiver_address, 10, initial_deposit=lambda x: 2 * x)
    channel.create_transfer(3)
    lost_balance_sig = channel.balance_sig
    channel.update_balance(0)

    with requests_mock.mock() as server_mock:
        headers1 = Munch()
        headers1.token_address = token_address
        headers1.contract_address = channel_manager_address
        headers1.receiver_address = receiver_address
        headers1.price = '7'

        headers2 = headers1.copy()
        headers2.invalid_amount = '1'
        headers2.sender_balance = '3'
        headers2.balance_signature = encode_hex(lost_balance_sig)

        headers3 = Munch()
        headers3.cost = '7'

        headers1 = HTTPHeaders.serialize(headers1)
        headers2 = HTTPHeaders.serialize(headers2)
        headers3 = HTTPHeaders.serialize(headers3)

        url = 'http://{}/something'.format(api_endpoint_address)
        server_mock.get(url, [
            {'status_code': 402, 'headers': headers1},
            {'status_code': 402, 'headers': headers2},
            {'status_code': 200, 'headers': headers3, 'text': 'success'}
        ])

        response = default_http_client.get(url)

    # First cycle, request price.
    request = server_mock.request_history[0]
    assert request.path == '/something'
    assert request.method == 'GET'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address

    # Second cycle, pay price based on outdated balance.
    request = server_mock.request_history[1]
    assert request.path == '/something'
    assert request.method == 'GET'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address
    assert request.headers['RDN-Balance'] == '7'
    assert request.headers['RDN-Balance-Signature']

    # Third cycle, adapt new balance and pay price again.
    request = server_mock.request_history[2]
    assert request.path == '/something'
    assert request.method == 'GET'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address
    assert request.headers['RDN-Balance'] == '10'

    channel = default_http_client.get_channel(url)
    assert channel.balance == 10
    balance_sig_hex = encode_hex(channel.balance_sig)
    assert request.headers['RDN-Balance-Signature'] == balance_sig_hex
    assert channel.balance_sig
    assert response.text == 'success'


def test_full_cycle_error_500(
        default_http_client: DefaultHTTPClient,
        api_endpoint_address: str,
        token_address: str,
        channel_manager_address: str,
        receiver_address: str
):
    default_http_client.initial_deposit = lambda x: x

    with requests_mock.mock() as server_mock:
        headers1 = Munch()
        headers1.token_address = token_address
        headers1.contract_address = channel_manager_address
        headers1.receiver_address = receiver_address
        headers1.price = '3'

        headers2 = Munch()
        headers2.cost = '3'

        headers1 = HTTPHeaders.serialize(headers1)
        headers2 = HTTPHeaders.serialize(headers2)

        url = 'http://{}/something'.format(api_endpoint_address)
        server_mock.get(url, [
            {'status_code': 402, 'headers': headers1},
            {'status_code': 500, 'headers': {}},
            {'status_code': 200, 'headers': headers2, 'text': 'success'}
        ])
        response = default_http_client.get(url)

    # First cycle, request price.
    request = server_mock.request_history[0]
    assert request.path == '/something'
    assert request.method == 'GET'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address

    # Second cycle, pay price but receive error.
    request = server_mock.request_history[1]
    assert request.path == '/something'
    assert request.method == 'GET'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address
    assert request.headers['RDN-Balance'] == '3'

    channel = default_http_client.get_channel(url)
    assert channel.balance == 3
    balance_sig_hex = encode_hex(channel.balance_sig)
    assert request.headers['RDN-Balance-Signature'] == balance_sig_hex

    # Third cycle, retry naively.
    request = server_mock.request_history[2]
    assert request.path == '/something'
    assert request.method == 'GET'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address
    assert request.headers['RDN-Balance'] == '3'
    assert channel.balance == 3
    assert request.headers['RDN-Balance-Signature'] == balance_sig_hex
    assert channel.balance_sig
    assert response.text == 'success'


def test_full_cycle_success_post(
        default_http_client: DefaultHTTPClient,
        api_endpoint_address: str,
        token_address: str,
        channel_manager_address: str,
        receiver_address: str
):
    default_http_client.initial_deposit = lambda x: x

    with requests_mock.mock() as server_mock:
        headers1 = Munch()
        headers1.token_address = token_address
        headers1.contract_address = channel_manager_address
        headers1.receiver_address = receiver_address
        headers1.price = '7'

        headers2 = Munch()
        headers2.cost = '7'

        headers1 = HTTPHeaders.serialize(headers1)
        headers2 = HTTPHeaders.serialize(headers2)

        url = 'http://{}/something'.format(api_endpoint_address)
        server_mock.post(url, [
            {'status_code': 402, 'headers': headers1},
            {'status_code': 200, 'headers': headers2, 'text': 'success'}
        ])
        response = default_http_client.post(url, json={'somefield': 'somevalue'})

    # First cycle, request price.
    request = server_mock.request_history[0]
    assert request.path == '/something'
    assert request.method == 'POST'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address
    assert request.json()['somefield'] == 'somevalue'

    # Second cycle, pay price.
    request = server_mock.request_history[1]
    assert request.path == '/something'
    assert request.method == 'POST'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address
    assert request.headers['RDN-Balance'] == '7'
    assert request.json()['somefield'] == 'somevalue'

    channel = default_http_client.get_channel(url)
    assert channel.balance == 7
    balance_sig_hex = encode_hex(channel.balance_sig)
    assert request.headers['RDN-Balance-Signature'] == balance_sig_hex
    assert channel.balance_sig
    assert response.text == 'success'


def test_custom_headers(
        default_http_client: DefaultHTTPClient,
        api_endpoint_address: str,
        token_address: str,
        channel_manager_address: str,
        receiver_address: str
):
    default_http_client.initial_deposit = lambda x: x

    with requests_mock.mock() as server_mock:
        headers1 = Munch()
        headers1.token_address = token_address
        headers1.contract_address = channel_manager_address
        headers1.receiver_address = receiver_address
        headers1.price = '7'

        headers2 = Munch()
        headers2.cost = '7'

        headers1 = HTTPHeaders.serialize(headers1)
        headers2 = HTTPHeaders.serialize(headers2)

        url = 'http://{}/something'.format(api_endpoint_address)
        server_mock.get(url, [
            {'status_code': 402, 'headers': headers1},
            {'status_code': 200, 'headers': headers2, 'text': 'success'}
        ])
        response = default_http_client.get(url, headers={
            'someheader': 'somevalue',
            # This should override the actual balance but doesn't actually make sense.
            'RDN-Balance': '5'
        })

    # First cycle, request price.
    request = server_mock.request_history[0]
    assert request.path == '/something'
    assert request.method == 'GET'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address
    assert request.headers['RDN-Balance'] == '5'
    assert request.headers['someheader'] == 'somevalue'

    # Second cycle, pay price.
    request = server_mock.request_history[1]
    assert request.path == '/something'
    assert request.method == 'GET'
    assert request.headers['RDN-Contract-Address'] == channel_manager_address
    assert request.headers['RDN-Balance'] == '5'
    assert request.headers['someheader'] == 'somevalue'

    channel = default_http_client.get_channel(url)
    assert channel.balance == 7
    balance_sig_hex = encode_hex(channel.balance_sig)
    assert request.headers['RDN-Balance-Signature'] == balance_sig_hex
    assert channel.balance_sig
    assert response.text == 'success'


def test_cheating_client(
        doggo_proxy,
        default_http_client: DefaultHTTPClient,
        http_doggo_url: str
):
    """this test scenario where client sends less funds than what is requested
        by the server. In such case, a "RDN-Invalid-Amount=1" header should
        be sent in a server's reply
    """
    def patched_payment(
            self: DefaultHTTPClient,
            method: str,
            url: str,
            response: Response,
            **kwargs
    ):
        response.headers[HTTPHeaders.PRICE] = str(
            int(response.headers[HTTPHeaders.PRICE]) + self.price_adjust
        )
        return DefaultHTTPClient.on_payment_requested(self, method, url, response, **kwargs)

    def patched_on_invalid_amount(self, method: str, url: str, response: Response, **kwargs):
        self.invalid_amount_received += 1
        price = int(response.headers[HTTPHeaders.PRICE])
        DefaultHTTPClient.on_invalid_amount(self, method, url, response, **kwargs)
        # on_invalid_amount will already prepare the next payment which we don't execute anymore,
        # so revert that.
        channel = self.get_channel(url)
        channel.update_balance(channel.balance - price)
        return False

    default_http_client.on_invalid_amount = types.MethodType(
        patched_on_invalid_amount,
        default_http_client
    )
    default_http_client.on_payment_requested = types.MethodType(
        patched_payment,
        default_http_client
    )

    default_http_client.invalid_amount_received = 0

    # correct amount
    default_http_client.price_adjust = 0
    response = default_http_client.get(http_doggo_url)
    check_response(response)
    assert default_http_client.invalid_amount_received == 0
    # underpay
    default_http_client.price_adjust = -1
    response = default_http_client.get(http_doggo_url)
    assert response is None
    assert default_http_client.invalid_amount_received == 1
    # overpay
    default_http_client.price_adjust = 1
    response = default_http_client.get(http_doggo_url)
    assert response is None
    assert default_http_client.invalid_amount_received == 2


def test_default_http_client(
        doggo_proxy,
        default_http_client: DefaultHTTPClient,
        sender_address: str,
        receiver_address: str,
        http_doggo_url: str
):
    check_response(default_http_client.get(http_doggo_url))

    client = default_http_client.client
    open_channels = client.get_open_channels()
    assert len(open_channels) == 1

    channel = open_channels[0]
    assert channel == default_http_client.get_channel(http_doggo_url)
    assert channel.balance_sig
    assert channel.balance < channel.deposit
    assert channel.sender == sender_address
    assert channel.receiver == receiver_address


def test_default_http_client_topup(
        doggo_proxy,
        default_http_client: DefaultHTTPClient,
        http_doggo_url: str
):

    # Create a channel that has just enough capacity for one transfer.
    default_http_client.initial_deposit = lambda x: 0
    check_response(default_http_client.get(http_doggo_url))

    client = default_http_client.client
    open_channels = client.get_open_channels()
    assert len(open_channels) == 1
    channel1 = open_channels[0]
    assert channel1 == default_http_client.get_channel(http_doggo_url)
    assert channel1.balance_sig
    assert channel1.balance == channel1.deposit

    # Do another payment. Topup should occur.
    check_response(default_http_client.get(http_doggo_url))
    open_channels = client.get_open_channels()
    assert len(open_channels) == 1
    channel2 = open_channels[0]
    assert channel2 == default_http_client.get_channel(http_doggo_url)
    assert channel2.balance_sig
    assert channel2.balance < channel2.deposit
    assert channel1 == channel2


def test_default_http_client_close(
        doggo_proxy,
        default_http_client: DefaultHTTPClient,
        http_doggo_url: str
):
    client = default_http_client.client
    check_response(default_http_client.get(http_doggo_url))
    default_http_client.close_active_channel(http_doggo_url)
    open_channels = client.get_open_channels()
    assert len(open_channels) == 0


def test_default_http_client_existing_channel(
        doggo_proxy,
        default_http_client: DefaultHTTPClient,
        receiver_address: str,
        http_doggo_url: str
):
    client = default_http_client.client
    channel = client.open_channel(receiver_address, 50)
    check_response(default_http_client.get(http_doggo_url))
    assert channel.balance == 2
    assert channel.deposit == 50


def test_default_http_client_existing_channel_topup(
        doggo_proxy,
        default_http_client: DefaultHTTPClient,
        receiver_address: str,
        http_doggo_url: str
):
    client = default_http_client.client
    default_http_client.topup_deposit = lambda x: 13
    channel = client.open_channel(receiver_address, 1)
    check_response(default_http_client.get(http_doggo_url))
    assert channel.balance == 2
    assert channel.deposit == 13


def test_coop_close(
        doggo_proxy,
        default_http_client: DefaultHTTPClient,
        http_doggo_url: str
):
    check_response(default_http_client.get(http_doggo_url))

    client = default_http_client.client
    open_channels = client.get_open_channels()
    assert len(open_channels) == 1

    channel = open_channels[0]
    import requests
    reply = requests.get('http://localhost:5000/api/1/channels/%s/%s' %
                         (channel.sender, channel.block))
    assert reply.status_code == 200
    json_reply = json.loads(reply.text)

    request_data = {'balance': json_reply['balance']}
    reply = requests.delete('http://localhost:5000/api/1/channels/%s/%s' %
                            (channel.sender, channel.block), data=request_data)

    assert reply.status_code == 200


@pytest.mark.parametrize('proxy_ssl', [1])
def test_ssl_client(
        doggo_proxy,
        default_http_client: DefaultHTTPClient,
        https_doggo_url: str
):
    with disable_ssl_check():
        check_response(default_http_client.get(https_doggo_url))
    with pytest.raises(SSLError):
        check_response(default_http_client.get(https_doggo_url))


def test_status_codes(
        doggo_proxy,
        default_http_client: DefaultHTTPClient,
        http_doggo_url: str
):
    patch_on_http_response(default_http_client, abort_on=[404])
    default_http_client.get(http_doggo_url)
    assert default_http_client.last_response.status_code == 200
    default_http_client.get(http_doggo_url[:-1])
    assert default_http_client.last_response.status_code == 404


def test_requests(
    init_microraiden_requests,
    revert_chain,
    api_endpoint_address: str,
    token_address: str,
    channel_manager_address: str,
    receiver_address: str
):
    import microraiden.requests

    with requests_mock.mock() as server_mock:
        headers1 = Munch()
        headers1.token_address = token_address
        headers1.contract_address = channel_manager_address
        headers1.receiver_address = receiver_address
        headers1.price = '7'

        headers2 = Munch()
        headers2.cost = '7'

        headers1 = HTTPHeaders.serialize(headers1)
        headers2 = HTTPHeaders.serialize(headers2)

        url = 'http://{}/something'.format(api_endpoint_address)
        server_mock.get(url, [
            {'status_code': 402, 'headers': headers1},
            {'status_code': 200, 'headers': headers2, 'text': 'success'}
        ])
        response = microraiden.requests.get(url)

    assert response.text == 'success'



import requests
from microraiden import HTTPHeaders as header


def test_accept_html(doggo_proxy, api_endpoint_address, wait_for_blocks):
    endpoint_url = "http://" + api_endpoint_address
    wait_for_blocks(1)

    headers = {
        'Accept': 'text/html'
    }
    rv = requests.get(endpoint_url + "/doggo.jpg", headers=headers)
    assert 'Content-Type' in rv.headers
    assert header.RECEIVER_ADDRESS in rv.headers
    assert header.CONTRACT_ADDRESS in rv.headers
    assert header.GATEWAY_PATH in rv.headers
    assert header.PRICE in rv.headers
    assert rv.headers['Content-Type'] == headers['Accept']
    assert rv.status_code == 402
    assert '<html' in rv.text


def test_accept_json(doggo_proxy, api_endpoint_address, wait_for_blocks):
    # default mode is to receive application/json
    endpoint_url = "http://" + api_endpoint_address
    wait_for_blocks(1)
    rv = requests.get(endpoint_url + "/doggo.jpg")
    assert 'Content-Type' in rv.headers
    assert header.RECEIVER_ADDRESS in rv.headers
    assert header.CONTRACT_ADDRESS in rv.headers
    assert header.GATEWAY_PATH in rv.headers
    assert header.PRICE in rv.headers
    assert rv.headers['Content-Type'] == 'application/json'
    assert rv.status_code == 402
    assert '<html' not in rv.text



from microraiden.test.fixtures import * # flake8: noqa
from gevent import monkey
monkey.patch_all(thread=False) # thread is false due to clash when testing both contract/microraiden modules
import logging

# to disable annoying 'test.rpc eth_getBlockNumber' message
logging.getLogger('testrpc.rpc').setLevel(logging.WARNING)


def pytest_addoption(parser):
    parser.addoption(
        "--no-tester",
        action="store_false",
        default=True,
        dest='use_tester',
        help="use a real RPC endpoint instead of the tester chain"
    )
    parser.addoption(
        "--no-clean-channels",
        action="store_false",
        default=True,
        dest='clean_channels',
        help="prevent all channels from closing cooperatively before and after each test"
    )



import gevent

from flask import Flask, safe_join
from flask_restful import (
    Api,
)

from microraiden import config

from microraiden.channel_manager import (
    ChannelManager
)

from microraiden.proxy.resources import (
    Expensive,
    ChannelManagementAdmin,
    ChannelManagementAdminChannels,
    ChannelManagementListChannels,
    ChannelManagementChannelInfo,
    ChannelManagementLogin,
    ChannelManagementLogout,
    ChannelManagementRoot,
    ChannelManagementStats,
)

from microraiden.proxy.resources.expensive import LightClientProxy
from microraiden.config import API_PATH
from microraiden.proxy.resources.paywall_decorator import Paywall


import logging
import ssl

log = logging.getLogger(__name__)


class PaywalledProxy:
    def __init__(
            self,
            channel_manager: ChannelManager,
            flask_app=None,
            paywall_html_dir=None,
            paywall_js_dir=None
    ):
        #  this doesn't work atm, due to test teardown problems
        #  it's not a critical error, but it should be fixed
        #  gevent.get_hub().SYSTEM_ERROR += (BaseException, )
        paywall_html_dir = paywall_html_dir or config.HTML_DIR
        paywall_js_dir = paywall_js_dir or config.JSLIB_DIR
        assert isinstance(channel_manager, ChannelManager)
        assert isinstance(paywall_html_dir, str)

        if not flask_app:
            self.app = Flask(
                __name__,
                static_url_path=config.JSPREFIX_URL,
                static_folder=paywall_js_dir
            )
        else:
            assert isinstance(flask_app, Flask)
            self.app = flask_app

        self.api = Api(self.app)
        self.rest_server = None
        self.server_greenlet = None

        self.channel_manager = channel_manager
        self.channel_manager.start()

        self.light_client_proxy = LightClientProxy(safe_join(paywall_html_dir, "index.html"))
        self.paywall = Paywall(channel_manager, self.light_client_proxy)

        # REST interface
        self.api.add_resource(ChannelManagementLogin, API_PATH + "/login")
        self.api.add_resource(ChannelManagementLogout, API_PATH + "/logout")
        self.api.add_resource(ChannelManagementChannelInfo,
                              API_PATH + "/channels/<string:sender_address>/<int:opening_block>",
                              resource_class_kwargs={'channel_manager': self.channel_manager})
        self.api.add_resource(ChannelManagementAdmin,
                              API_PATH + "/admin",
                              resource_class_kwargs={'channel_manager': self.channel_manager})
        self.api.add_resource(ChannelManagementAdminChannels,
                              API_PATH +
                              "/admin/channels/<string:sender_address>/<int:opening_block>",
                              resource_class_kwargs={'channel_manager': self.channel_manager})
        self.api.add_resource(ChannelManagementListChannels,
                              API_PATH + "/channels/",
                              API_PATH + "/channels/<string:sender_address>",
                              resource_class_kwargs={'channel_manager': self.channel_manager})
        self.api.add_resource(ChannelManagementStats,
                              API_PATH + "/stats",
                              resource_class_kwargs={'channel_manager': self.channel_manager})
        self.api.add_resource(ChannelManagementRoot, "/cm")

    def run(self, host='localhost', port=5000, debug=False, ssl_context=None):
        assert ssl_context is None or len(ssl_context) == 2
        # register our custom error handler to ignore some exceptions and fail on others
#        register_error_handler(self.gevent_error_handler)
        self.channel_manager.wait_sync()
        from gevent.pywsgi import WSGIServer
        if ((ssl_context is not None) and
           (len(ssl_context) == 2) and
           (ssl_context[0] and ssl_context[1])):
            self.rest_server = WSGIServer((host, port), self.app,
                                          keyfile=ssl_context[0],
                                          certfile=ssl_context[1])
        else:
            self.rest_server = WSGIServer((host, port), self.app)
        self.server_greenlet = gevent.spawn(self.rest_server.serve_forever)

    def stop(self):
        assert self.rest_server is not None
        assert self.server_greenlet is not None
        # we should stop the server only if it has been started. In case we do stop()
        #  right after start(), the server may be in an undefined state and join() will
        #  hang indefinetely (this often happens with tests)
        for try_n in range(5):
            if self.rest_server.started is True:
                break
            gevent.sleep(1)
        self.rest_server.stop()
        self.server_greenlet.join()

    def join(self):
        try:
            self.server_greenlet.join()
        finally:
            self.channel_manager.stop()

    @staticmethod
    def gevent_error_handler(context, exc_info):
        e = exc_info[1]
        # recover if HTTP request is done to a HTTPS endpoint
        if isinstance(e, ssl.SSLError) and e.reason == 'HTTP_REQUEST':
            return
        gevent.get_hub().handle_system_error(exc_info[0], exc_info[1])

    def add_paywalled_resource(self, cls: Expensive, url: str, price: int=None, *args, **kwargs):
        cfg = {
            'channel_manager': self.channel_manager,
            'light_client_proxy': self.light_client_proxy,
            'price': price,
            'paywall': self.paywall
        }
        if 'resource_class_kwargs' in kwargs:
            cfg.update(kwargs.pop('resource_class_kwargs'))
        return self.api.add_resource(cls, url, resource_class_kwargs=cfg, *args, **kwargs)



from .crypto import (
    generate_privkey,
    pubkey_to_addr,
    privkey_to_addr,
    addr_from_sig,
    pack,
    keccak256,
    keccak256_hex,
    sign,
    sign_transaction,
    eth_message_hash,
    eth_sign,
    eth_verify,
    eth_sign_typed_data_message,
    eth_sign_typed_data,
    eth_sign_typed_data_message_eip,
    eth_sign_typed_data_eip,
    get_balance_message,
    sign_balance_proof,
    verify_balance_proof,
    sign_close,
    verify_closing_sig
)

from .contract import (
    create_signed_transaction,
    create_transaction,
    create_signed_contract_transaction,
    create_contract_transaction,
    create_transaction_data,
    get_logs,
    get_event_blocking,
    wait_for_transaction
)

from .private_key import (
    check_permission_safety,
    get_private_key
)

__all__ = [
    generate_privkey,
    pubkey_to_addr,
    privkey_to_addr,
    addr_from_sig,
    pack,
    keccak256,
    keccak256_hex,
    sign,
    sign_transaction,
    eth_message_hash,
    eth_sign,
    eth_verify,
    eth_sign_typed_data_message,
    eth_sign_typed_data,
    eth_sign_typed_data_message_eip,
    eth_sign_typed_data_eip,
    get_balance_message,
    sign_balance_proof,
    verify_balance_proof,
    sign_close,
    verify_closing_sig,

    create_signed_transaction,
    create_transaction,
    create_signed_contract_transaction,
    create_contract_transaction,
    create_transaction_data,
    get_logs,
    get_event_blocking,
    wait_for_transaction,

    check_permission_safety,
    get_private_key
]



import logging
from typing import Tuple, Any

import requests
from eth_utils import encode_hex, decode_hex, is_same_address
from munch import Munch
from requests import Response
from urllib.parse import urlparse

from microraiden.client import Channel
from microraiden.header import HTTPHeaders
from .client import Client

log = logging.getLogger(__name__)


class HTTPClient(object):
    def __init__(self, client: Client) -> None:
        self.client = client
        self.endpoint_to_channel = {}

    def request(self, method: str, url: str, **kwargs) -> Response:
        self.on_init(method, url, **kwargs)
        retry = True
        response = None
        while retry:
            response, retry = self._request_resource(method, url, **kwargs)

        self.on_exit(method, url, response, **kwargs)
        return response

    def head(self, url: str, **kwargs) -> Response:
        return self.request('HEAD', url, **kwargs)

    def get(self, url: str, **kwargs) -> Response:
        return self.request('GET', url, **kwargs)

    def post(self, url: str, **kwargs) -> Response:
        return self.request('POST', url, **kwargs)

    def put(self, url: str, **kwargs) -> Response:
        return self.request('PUT', url, **kwargs)

    def patch(self, url: str, **kwargs) -> Response:
        return self.request('PATCH', url, **kwargs)

    def delete(self, url: str, **kwargs) -> Response:
        return self.request('DELETE', url, **kwargs)

    @staticmethod
    def get_endpoint(url: str) -> str:
        url_parts = urlparse(url)
        assert url_parts.scheme, 'No protocol scheme specified.'
        return '://'.join(url_parts[:2])

    @staticmethod
    def close_channel(endpoint_url: str, channel: Channel):
        log.debug(
            'Requesting closing signature from server for balance {} on channel {}/{}/{}.'
            .format(channel.balance, channel.sender, channel.sender, channel.block)
        )
        url = '{}/api/1/channels/{}/{}'.format(endpoint_url, channel.sender, channel.block)
        response = requests.delete(url, data={'balance': channel.balance})
        if response.status_code == requests.codes.OK:
            closing_sig = response.json()['close_signature']
            channel.close_cooperatively(decode_hex(closing_sig))
        else:
            log.error('No closing signature received: {}'.format(response.text))

    def get_channel(self, url: str) -> Channel:
        return self.endpoint_to_channel.get(self.get_endpoint(url))

    def set_channel(self, url: str, channel: Channel):
        self.endpoint_to_channel[self.get_endpoint(url)] = channel

    def close_active_channel(self, url: str):
        endpoint = self.get_endpoint(url)
        channel = self.get_channel(url)
        if channel:
            self.close_channel(endpoint, channel)

    def _request_resource(self, method: str, url: str, **kwargs) -> Tuple[Any, bool]:
        """
        Performs a simple GET request to the HTTP server with headers representing the given
        channel state.
        """
        headers = Munch()
        headers.contract_address = self.client.context.channel_manager.address
        channel = self.get_channel(url)
        if channel:
            headers.balance = str(channel.balance)
            headers.balance_signature = encode_hex(channel.balance_sig)
            headers.sender_address = channel.sender
            headers.receiver_address = channel.receiver
            headers.open_block = str(channel.block)

        headers = HTTPHeaders.serialize(headers)
        if 'headers' in kwargs:
            headers.update(kwargs['headers'])
            kwargs['headers'] = headers
        else:
            kwargs['headers'] = headers
        response = requests.request(method, url, **kwargs)
        response_headers = HTTPHeaders.deserialize(response.headers)

        if self.on_http_response(method, url, response, **kwargs) is False:
            return None, False  # user requested abort

        if response.status_code == requests.codes.OK:
            return response, self.on_success(method, url, response, **kwargs)

        elif response.status_code == requests.codes.PAYMENT_REQUIRED:
            if 'insuf_confs' in response_headers:
                return None, self.on_insufficient_confirmations(method, url, response, **kwargs)

            elif 'insuf_funds' in response_headers:
                return None, self.on_insufficient_funds(method, url, response, **kwargs)

            elif 'contract_address' not in response_headers or not is_same_address(
                response_headers.contract_address,
                self.client.context.channel_manager.address
            ):
                return None, self.on_invalid_contract_address(method, url, response, **kwargs)

            elif 'invalid_amount' in response_headers:
                return None, self.on_invalid_amount(method, url, response, **kwargs)

            else:
                return None, self.on_payment_requested(method, url, response, **kwargs)
        else:
            return None, self.on_http_error(method, url, response, **kwargs)

    def on_init(self, method: str, url: str, **kwargs):
        log.debug('Starting {} request loop for resource at {}.'.format(method, url))

    def on_exit(self, method: str, url: str, response: Response, **kwargs):
        pass

    def on_success(self, method: str, url: str, response: Response, **kwargs) -> bool:
        log.debug('Resource received.')
        cost = response.headers.get(HTTPHeaders.COST)
        if cost is not None:
            log.debug('Final cost was {}.'.format(cost))
        return False

    def on_insufficient_funds(self, method: str, url: str, response: Response, **kwargs) -> bool:
        log.error(
            'Server was unable to verify the transfer - Insufficient funds of the balance proof '
            'or possibly an unconfirmed or unregistered topup.'
        )
        return False

    def on_insufficient_confirmations(
            self,
            method: str,
            url: str,
            response: Response,
            **kwargs
    ) -> bool:
        return True

    def on_invalid_amount(self, method: str, url: str, response: Response, **kwargs) -> bool:
        return True

    def on_invalid_contract_address(
            self,
            method: str,
            url: str,
            response: Response,
            **kwargs
    ) -> bool:
        contract_address = response.headers.get(HTTPHeaders.CONTRACT_ADDRESS)
        log.error(
            'Server sent no or invalid contract address: {}.'.format(contract_address)
        )
        return False

    def on_payment_requested(self, method: str, url: str, response: Response, **kwargs):
        return True

    def on_http_response(self, method: str, url: str, response: Response, **kwargs):
        """Called whenever server returns a reply.
        Return False to abort current request."""
        log.debug('Response received: {}'.format(response.headers))
        return True

    def on_http_error(self, method: str, url: str, response: Response, **kwargs):
        """Triggered under the default behavior when the server returns anything other than a 402
        or 200."""
        return False



import logging

from microraiden.channel_manager import (
    ChannelManager,
)

from flask_restful import Resource
from eth_utils import is_address
from .paywall_decorator import paywall_decorator

log = logging.getLogger(__name__)


class LightClientProxy:
    def __init__(self, index_html):
        with open(index_html) as fp:
            self.data = fp.read()

    def get(self, url):
        return self.data


class Expensive(Resource):
    method_decorators = [paywall_decorator]

    def __init__(self,
                 channel_manager: ChannelManager,
                 light_client_proxy=None,
                 paywall=None,
                 price: None = None,
                 ) -> None:
        super(Expensive, self).__init__()
        assert isinstance(channel_manager, ChannelManager)
        assert price is None or callable(price) or price > 0
        self.contract_address = channel_manager.channel_manager_contract.address
        self.receiver_address = channel_manager.receiver
        assert is_address(self.contract_address)
        assert is_address(self.receiver_address)
        self.channel_manager = channel_manager
        self.light_client_proxy = light_client_proxy
        self._price = price
        self.paywall = paywall

    def get_paywall(self, url):
        return self.light_client_proxy.get(url)

    def price(self):
        if callable(self._price):
            return self._price()
        else:
            return self._price



import pytest # noqa
import requests
import json
import gevent

from microraiden.config import API_PATH
from microraiden.proxy.paywalled_proxy import PaywalledProxy


def test_resources(doggo_proxy, api_endpoint_address, users_db):
    auth_credentials = ('user', 'password')
    users_db.add_user(auth_credentials[0], auth_credentials[1])
    users_db.token_expiry_seconds = 0.5
    api_path = "http://" + api_endpoint_address + API_PATH

    # without auth, refuse access
    rv = requests.get(api_path + "/admin")
    assert rv.status_code == 401

    #  bad login
    rv = requests.get(api_path + "/login", auth=('user', 'bad_password'))
    assert rv.status_code == 401

    # good login, we got the token
    rv = requests.get(api_path + "/login", auth=auth_credentials)
    assert rv.status_code == 200
    json_response = json.loads(rv.text)
    assert 'token' in json_response

    token_credentials = (json_response['token'], '')
    # use the token to login
    rv = requests.get(api_path + "/admin", auth=token_credentials)
    assert rv.status_code == 200

    # logout with an invalid token
    rv = requests.get(api_path + "/logout", auth=('invalid_token', ''))
    assert rv.status_code == 401

    # logout with a valid token
    rv = requests.get(api_path + "/logout", auth=token_credentials)
    assert rv.status_code == 200

    # after logout, refuse access
    rv = requests.get(api_path + "/admin", auth=token_credentials)
    assert rv.status_code == 401

    # TODO: test token expiration. we must set token expiry timeout somehow
    # login again
    rv = requests.get(api_path + "/login", auth=auth_credentials)
    assert rv.status_code == 200
    json_response = json.loads(rv.text)
    assert 'token' in json_response

    token_credentials = (json_response['token'], '')
    gevent.sleep(1)

    # use the token to login
    rv = requests.get(api_path + "/admin", auth=token_credentials)
    assert rv.status_code == 401


def test_stats(doggo_proxy: PaywalledProxy, api_endpoint_address):
    api_path = "http://" + api_endpoint_address + API_PATH

    rv = requests.get(api_path + "/stats")
    assert rv.status_code == 200
    stats = json.loads(rv.text)
    assert 'balance_sum' in stats
    assert 'deposit_sum' in stats
    assert 'token_address' in stats
    assert 'receiver_address' in stats
    assert 'contract_address' in stats
    assert stats['receiver_address'] == doggo_proxy.channel_manager.receiver
    token_address = doggo_proxy.channel_manager.token_contract.address
    assert stats['token_address'] == token_address
    contract_address = doggo_proxy.channel_manager.channel_manager_contract.address
    assert stats['contract_address'] == contract_address



import pytest
from web3 import Web3
from web3.contract import Contract

from microraiden.test.config import (
    RECEIVER_ETH_ALLOWANCE,
    RECEIVER_TOKEN_ALLOWANCE
)
from microraiden.channel_manager import ChannelManager
from microraiden.exceptions import InvalidContractVersion
from microraiden.utils import privkey_to_addr


class FakeChannelManagerContract:
    def __init__(self, receiver1_privkey):
        self._address = privkey_to_addr(receiver1_privkey)

    def call(self):
        # forgive me God for I have sinned here
        return self

    def version(self):
        return "6.6.6"

    @property
    def address(self):
        return self._address


def test_version(
        web3: Web3,
        channel_manager_contract: Contract,
        token_contract: Contract,
        make_account
):
    """Test if proxy refuses to run if it deployed contract version
    is different from the one it expects"""
    receiver1_privkey = make_account(RECEIVER_ETH_ALLOWANCE, RECEIVER_TOKEN_ALLOWANCE)

    # this one should not raise
    cm = ChannelManager(
        web3,
        channel_manager_contract,
        token_contract,
        receiver1_privkey,
        state_filename=":memory:"
    )
    cm.stop()
    # now we patch it
    channel_manager_contract = FakeChannelManagerContract(receiver1_privkey)

    # check of version string should fail here
    with pytest.raises(InvalidContractVersion):
        ChannelManager(
            web3,
            channel_manager_contract,
            token_contract,
            receiver1_privkey,
            state_filename=":memory:"
        )



import pytest
import random

from eth_utils import (
    decode_hex,
    encode_hex,
    int_to_big_endian,
    pad_left,
)
import ethereum.tester
from web3 import Web3
from web3.contract import Contract

from microraiden.utils import (
    privkey_to_addr,
    create_signed_contract_transaction,
    create_signed_transaction
)
from microraiden.config import GAS_PRICE
from microraiden.test.config import (
    RECEIVER_ETH_ALLOWANCE,
    RECEIVER_TOKEN_ALLOWANCE,
    SENDER_ETH_ALLOWANCE,
    SENDER_TOKEN_ALLOWANCE,
    FAUCET_PRIVKEY,
    FAUCET_ADDRESS,
)


def random_private_key(bound):
    """Randomly gnerate a private key smaller than a certain bound."""
    n = random.randint(1, bound)  # nosec
    private_key = encode_hex(pad_left(int_to_big_endian(n), 32, '\0'))
    return private_key


@pytest.fixture(scope='session')
def make_account(
        request,
        web3: Web3,
        wait_for_transaction,
        use_tester: bool,
        token_contract: Contract
):
    def account_factory(eth_allowance, token_allowance):
        privkey = random_private_key(1000)
        address = privkey_to_addr(privkey)
        if use_tester:
            ethereum.tester.accounts.append(decode_hex(address))
            ethereum.tester.keys.append(decode_hex(privkey))
        fund_account(
            address,
            eth_allowance,
            token_allowance,
            token_contract,
            web3,
            wait_for_transaction
        )

        def finalize():
            sweep_account(privkey, token_contract, web3, wait_for_transaction)
            if use_tester:
                ethereum.tester.accounts.remove(decode_hex(address))
                ethereum.tester.keys.remove(decode_hex(privkey))
        request.addfinalizer(finalize)
        return privkey
    return account_factory


def fund_account(
        address: str,
        eth_allowance: int,
        token_allowance: int,
        token_contract: Contract,
        web3: Web3,
        wait_for_transaction
):
    tx = create_signed_transaction(
        FAUCET_PRIVKEY,
        web3,
        to=address,
        value=eth_allowance,
        gas_limit=21000
    )
    tx_hash = web3.eth.sendRawTransaction(tx)
    wait_for_transaction(tx_hash)

    if token_allowance > 0:
        tx = create_signed_contract_transaction(
            FAUCET_PRIVKEY,
            token_contract,
            'transfer',
            [
                address,
                token_allowance
            ]
        )
        tx_hash = web3.eth.sendRawTransaction(tx)
        wait_for_transaction(tx_hash)


def sweep_account(private_key: str, token_contract: Contract, web3: Web3, wait_for_transaction):
    address = privkey_to_addr(private_key)
    token_balance = token_contract.call().balanceOf(address)
    if token_balance > 0:
        tx = create_signed_contract_transaction(
            private_key,
            token_contract,
            'transfer',
            [
                FAUCET_ADDRESS,
                token_balance
            ]
        )
        try:
            tx_hash = web3.eth.sendRawTransaction(tx)
        except ValueError as e:
            if e.args[0]['message'].startswith('Insufficient funds.'):
                pass
            else:
                raise
        else:
            wait_for_transaction(tx_hash)

    balance = web3.eth.getBalance(address, 'pending')
    if balance < 21000 * GAS_PRICE:
        return
    tx = create_signed_transaction(
        private_key,
        web3,
        to=FAUCET_ADDRESS,
        value=balance - 21000 * GAS_PRICE,
        gas_limit=21000
    )
    tx_hash = web3.eth.sendRawTransaction(tx)
    wait_for_transaction(tx_hash)
    assert web3.eth.getBalance(address, 'pending') == 0, (
        'Sweeping of account {} (private key {}) failed.'.format(address, private_key)
    )


@pytest.fixture
def sender_privkey(make_account):
    return make_account(SENDER_ETH_ALLOWANCE, SENDER_TOKEN_ALLOWANCE)


@pytest.fixture
def sender_address(sender_privkey):
    return privkey_to_addr(sender_privkey)


@pytest.fixture
def receiver_privkey(make_account):
    return make_account(RECEIVER_ETH_ALLOWANCE, RECEIVER_TOKEN_ALLOWANCE)


@pytest.fixture
def receiver_address(receiver_privkey):
    return privkey_to_addr(receiver_privkey)



"""
This is dummy code showing how the minimal app could look like.
In his case we don't use a proxy, but directly a server
"""
import os
import click
from microraiden.proxy.content import (
    PaywalledContent,
    PaywalledProxyUrl
)
from microraiden.make_helpers import make_paywalled_proxy
from microraiden.config import TKN_DECIMALS

if __name__ == '__main__':
    private_key = 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
    state_file_name = 'echo_server.json'
    app_dir = click.get_app_dir('microraiden')
    if not os.path.exists(app_dir):
        os.makedirs(app_dir)
    state_file_name = os.path.join(app_dir, state_file_name)
    # set up a paywalled proxy
    # arguments are:
    #  - private key to use for receiving funds
    #  - file for storing state information (balance proofs)
    app = make_paywalled_proxy(private_key, state_file_name)

    # Add resource defined by regex and with a fixed price of 1 token
    # We setup the resource to return whatever is supplied as a second argument
    #  in the URL.
    app.add_content(PaywalledContent(
                    "echofix\/[a-zA-Z0-9]+", 1 * TKN_DECIMALS,
                    lambda request: (request.split("/")[1], 200)))
    # Resource with a price determined by the second parameter
    app.add_content(PaywalledContent(
                    "echodyn\/[0-9]+",
                    lambda request: int(request.split("/")[1]) * TKN_DECIMALS,
                    lambda request: (int(request.split("/")[1]), 200)))
    app.add_content(PaywalledProxyUrl(
                    "p\/[0-9]+",
                    1 * TKN_DECIMALS,
                    lambda request: 'google.com/search?q=' + request.split("/")[1]))
    # start the app. proxy is a WSGI greenlet, so you must join it properly
    app.run(debug=True)
    app.join()
    # now use echo_client to get the resources



import pytest
import types
from microraiden.channel_manager import ChannelManager, Blockchain


def start_channel_manager(channel_manager, use_tester, mine_sync_event):
    if use_tester:
        mine_sync_event.clear()

        # monkeypatch Blockchain::_update() to wait for an sync event
        def update_patched(self: Blockchain):
            Blockchain._update(self)
            mine_sync_event.wait()

        channel_manager.blockchain._update = types.MethodType(
            update_patched, channel_manager.blockchain)
        # it is pointless to do busy loop as the Blockchain blocks on sync
        channel_manager.blockchain.poll_interval = 0

        def stop_patched(self: ChannelManager):
            mine_sync_event.set()
            ChannelManager.stop(self)
            self.stop = types.MethodType(ChannelManager.stop, self)

        channel_manager.stop = types.MethodType(
            stop_patched, channel_manager)

    def fail(greenlet):
        raise greenlet.exception

    channel_manager.link_exception(fail)
    channel_manager.start()
    return channel_manager


@pytest.fixture
def channel_manager(
        web3,
        receiver_privkey,
        channel_manager_contract,
        token_contract,
        use_tester,
        mine_sync_event,
        state_db_path,
        patched_contract,
        revert_chain
):
    manager = ChannelManager(
        web3,
        channel_manager_contract,
        token_contract,
        receiver_privkey,
        n_confirmations=5,
        state_filename=state_db_path
    )
    start_channel_manager(manager, use_tester, mine_sync_event)
    yield manager
    manager.stop()



from web3 import Web3

from microraiden import Client
from microraiden.client import Channel
from microraiden.utils import sign_balance_proof, sign_close
from microraiden.test.utils.client import close_channel_cooperatively


def test_client(client: Client, receiver_address):
    """test if contract calls go through"""

    c = client.open_channel(receiver_address, 10)
    assert c is not None

    sig = c.create_transfer(5)
    assert sig is not None

    ev = c.topup(10)
    assert ev is not None
    assert c.deposit == 20

    ev = c.close()
    assert ev is not None


def test_cooperative_close(client: Client, receiver_privkey, receiver_address):
    c = client.get_suitable_channel(receiver_address, 3)
    c.create_transfer(3)

    assert c.deposit >= 3
    assert c.balance == 3

    sig = sign_close(receiver_privkey, c.balance_sig)
    assert c.close_cooperatively(sig)
    assert c.state == Channel.State.closed


def test_integrity(client: Client, receiver_address):
    c = client.get_suitable_channel(receiver_address, 5)
    assert c.balance == 0
    assert c.balance_sig == sign_balance_proof(
        client.context.private_key,
        receiver_address,
        c.block,
        0,
        client.context.channel_manager.address
    )
    assert c.is_valid()

    # Balance update without sig update.
    c._balance = 2
    assert not c.is_valid()

    # Proper balance update with sig update.
    c.update_balance(2)
    assert c.is_valid()

    # Random sig.
    c._balance_sig = b'wrong'
    assert not c.is_valid()

    # Balance exceeds deposit.
    c.update_balance(100)
    assert not c.is_valid()


def test_sync(client: Client, receiver_address, receiver_privkey):
    c = client.get_suitable_channel(receiver_address, 5, initial_deposit=lambda x: x)
    assert c in client.channels
    assert c.deposit == 5
    assert len(client.channels) == 1

    # Check if channel is still valid after sync.
    client.sync_channels()
    assert c in client.channels
    assert len(client.channels) == 1

    # Check if client handles topup events on sync.
    c_topup = client.get_suitable_channel(receiver_address, 7, topup_deposit=lambda x: 2)
    assert c_topup == c
    assert len(client.channels) == 1
    assert c.deposit == 7

    # Check if channel can be resynced after data loss.
    client.channels = []
    client.sync_channels()
    assert len(client.channels) == 1
    c = client.channels[0]
    assert c.deposit == 7

    # Check if channel is forgotten on resync after closure.
    close_channel_cooperatively(c, receiver_privkey)

    client.sync_channels()
    assert c not in client.channels


def test_open_channel_insufficient_tokens(client: Client, web3: Web3, receiver_address: str):
    balance_of = client.context.token.call().balanceOf(client.context.address)
    tx_count_pre = web3.eth.getTransactionCount(client.context.address)
    channel = client.open_channel(receiver_address, balance_of + 1)
    tx_count_post = web3.eth.getTransactionCount(client.context.address)
    assert channel is None
    assert tx_count_post == tx_count_pre


def test_topup_channel_insufficient_tokens(client: Client, web3: Web3, receiver_address: str):
    balance_of = client.context.token.call().balanceOf(client.context.address)
    channel = client.open_channel(receiver_address, 1)

    tx_count_pre = web3.eth.getTransactionCount(client.context.address)
    assert channel.topup(balance_of) is None
    tx_count_post = web3.eth.getTransactionCount(client.context.address)
    assert tx_count_post == tx_count_pre



import os
from eth_utils import denoms
from microraiden.utils import privkey_to_addr

CHANNEL_MANAGER_ADDRESS = '0xeb244b0502a2d3867e5cab2347c6e1cdeb5e1eef'
API_PATH = "/api/1"

MICRORAIDEN_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), '../..'))
HTML_DIR = os.path.join(MICRORAIDEN_DIR, 'microraiden', 'webui')
JSLIB_DIR = os.path.join(HTML_DIR, 'js')

FAUCET_PRIVKEY = 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa'
FAUCET_ADDRESS = privkey_to_addr(FAUCET_PRIVKEY)
FAUCET_ALLOWANCE = 10**23
INITIAL_TOKEN_SUPPLY = 10**25
SENDER_ETH_ALLOWANCE = int(0.02 * denoms.ether)
SENDER_TOKEN_ALLOWANCE = 10**20
RECEIVER_ETH_ALLOWANCE = int(0.02 * denoms.ether)
RECEIVER_TOKEN_ALLOWANCE = 0



from .expensive import Expensive
from .management import (
    ChannelManagementRoot,
    ChannelManagementAdmin,
    ChannelManagementAdminChannels,
    ChannelManagementListChannels,
    ChannelManagementStats,
    ChannelManagementChannelInfo,
)
from .login import (
    ChannelManagementLogin,
    ChannelManagementLogout,
)
from .proxy_url import PaywalledProxyUrl

__all__ = (
    Expensive,
    ChannelManagementRoot,
    ChannelManagementListChannels,
    ChannelManagementChannelInfo,
    ChannelManagementAdmin,
    ChannelManagementAdminChannels,
    ChannelManagementStats,
    ChannelManagementLogin,
    ChannelManagementLogout,
    PaywalledProxyUrl
)



import pytest

from microraiden.channel_manager import ChannelManager
from microraiden.proxy.paywalled_proxy import PaywalledProxy
from microraiden.proxy.resources import Expensive
import microraiden.proxy.resources.login as login
import logging
log = logging.getLogger(__name__)


class ExpensiveKitten(Expensive):
    def get(self, path):
        return "HI I AM A KITTEN"


class ExpensiveDoggo(Expensive):
    def get(self, path):
        return "HI I AM A DOGGO"


class ExpensiveTeapot(Expensive):
    def get(self, path):
        return "HI I AM A TEAPOT", 418


@pytest.fixture
def proxy_state_filename():
    return None


@pytest.fixture
def empty_proxy(channel_manager: ChannelManager, wait_for_blocks, use_tester: bool):
    app = PaywalledProxy(channel_manager)
    app.run()

    if use_tester:
        # Waiting only required on tester chains due to monkey-patched channel manager.
        wait_for_blocks(1)

    yield app
    app.stop()


@pytest.fixture
def doggo_proxy(
        channel_manager: ChannelManager,
        proxy_ssl: bool,
        proxy_ssl_certs
):
    app = PaywalledProxy(channel_manager)
    app.add_paywalled_resource(ExpensiveKitten, '/kitten.jpg', 1)
    app.add_paywalled_resource(ExpensiveDoggo, '/doggo.jpg', 2)
    app.add_paywalled_resource(ExpensiveTeapot, '/teapot.jpg', 3)
    ssl_context = proxy_ssl_certs if proxy_ssl else None
    app.run(ssl_context=ssl_context)
    yield app
    app.stop()


@pytest.fixture
def http_doggo_url(api_endpoint_address) -> str:
    return 'http://{}/doggo.jpg'.format(api_endpoint_address)


@pytest.fixture
def https_doggo_url(api_endpoint_address) -> str:
    return 'https://{}/doggo.jpg'.format(api_endpoint_address)


@pytest.fixture
def users_db():
    return login.userDB



from gevent.hub import Hub


IGNORE_ERROR = Hub.SYSTEM_ERROR + Hub.NOT_ERROR


def register_error_handler(error_handler):

    Hub._origin_handle_error = Hub.handle_error

    def custom_handle_error(self, context, type, value, tb):
        if not issubclass(type, IGNORE_ERROR):
            # print 'Got error from greenlet:', context, type, value, tb
            error_handler(context, (type, value, tb))

        self._origin_handle_error(context, type, value, tb)

    Hub.handle_error = custom_handle_error



import sys

from web3 import Web3, HTTPProvider

import logging

from web3.contract import Contract

log = logging.getLogger(__name__)

from microraiden.channel_manager import ChannelManager
from microraiden.exceptions import (
    StateReceiverAddrMismatch,
    StateContractAddrMismatch
)
from microraiden import config
from microraiden.proxy.paywalled_proxy import PaywalledProxy


def make_channel_manager_contract(web3: Web3, channel_manager_address: str) -> Contract:
    return web3.eth.contract(
        abi=config.CONTRACT_METADATA[config.CHANNEL_MANAGER_ABI_NAME]['abi'],
        address=channel_manager_address
    )


def make_channel_manager(
        private_key: str,
        channel_manager_address: str,
        state_filename: str,
        web3: Web3
) -> ChannelManager:
    channel_manager_contract = make_channel_manager_contract(web3, channel_manager_address)
    token_address = channel_manager_contract.call().token()
    token_abi = config.CONTRACT_METADATA[config.TOKEN_ABI_NAME]['abi']
    token_contract = web3.eth.contract(abi=token_abi, address=token_address)
    try:
        return ChannelManager(
            web3,
            channel_manager_contract,
            token_contract,
            private_key,
            state_filename=state_filename
        )
    except StateReceiverAddrMismatch as e:
        log.error(
            'Receiver address does not match address stored in a saved state. '
            'Use a different file, or backup and remove %s. (%s)' %
            (state_filename, e)
        )
        sys.exit(1)

    except StateContractAddrMismatch as e:
        log.error(
            'Channel contract address mismatch. '
            'Saved state file is %s. Backup it, remove, then start proxy again (%s)' %
            (state_filename, e)
        )
        sys.exit(1)


def make_paywalled_proxy(
        private_key: str,
        state_filename: str,
        contract_address=config.CHANNEL_MANAGER_ADDRESS,
        flask_app=None,
        web3=None
) -> PaywalledProxy:
    if web3 is None:
        web3 = Web3(HTTPProvider(config.WEB3_PROVIDER_DEFAULT, request_kwargs={'timeout': 60}))
    channel_manager = make_channel_manager(private_key, contract_address, state_filename, web3)
    proxy = PaywalledProxy(channel_manager, flask_app, config.HTML_DIR, config.JSLIB_DIR)
    return proxy



import logging
from itertools import count
import os
import sys
import gevent
import traceback

import click
from eth_utils import (
    decode_hex,
    is_same_address,
)
from ethereum.tester import TransactionFailed
from web3 import Web3, HTTPProvider, RPCProvider
from web3.contract import Contract
from web3.exceptions import BadFunctionCallOutput

from microraiden import (
    config,
    utils,
)
from microraiden.channel_manager import ChannelManagerState
from microraiden.utils import create_signed_contract_transaction, privkey_to_addr
from microraiden.exceptions import StateFileException
from microraiden.make_helpers import make_channel_manager_contract

log = logging.getLogger('close_all_channels')


@click.command()
@click.option(
    '--rpc-provider',
    default=config.WEB3_PROVIDER_DEFAULT,
    help='Address of the Ethereum RPC provider'
)
@click.option(
    '--private-key',
    required=True,
    help='Path to private key file of the proxy',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    '--private-key-password-file',
    default=None,
    help='Path to file containing password for the JSON-encoded private key',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    '--state-file',
    default=None,
    help='State file of the proxy'
)
@click.option(
    '--channel-manager-address',
    default=None,
    help='Ethereum address of the channel manager contract'
)
def main(
        rpc_provider: RPCProvider,
        private_key: str,
        private_key_password_file: str,
        state_file: str,
        channel_manager_address: str
):
    private_key = utils.get_private_key(private_key, private_key_password_file)
    if private_key is None:
        sys.exit(1)

    receiver_address = privkey_to_addr(private_key)
    channel_manager_address = channel_manager_address or config.CHANNEL_MANAGER_ADDRESS

    if not state_file:
        state_file_name = "%s_%s.json" % (channel_manager_address[:10], receiver_address[:10])
        app_dir = click.get_app_dir('microraiden')
        state_file = os.path.join(app_dir, state_file_name)

    web3 = Web3(HTTPProvider(rpc_provider, request_kwargs={'timeout': 60}))
    web3.eth.defaultAccount = receiver_address
    channel_manager_contract = make_channel_manager_contract(web3, config.CHANNEL_MANAGER_ADDRESS)

    try:
        click.echo('Loading state file from {}'.format(state_file))
        state = ChannelManagerState.load(state_file)
    except StateFileException:
        click.echo('Error reading state file')
        traceback.print_exc()
        sys.exit(1)
    if not is_same_address(state.receiver, receiver_address):
        click.echo('Private key does not match receiver address in state file')
        sys.exit(1)
    elif not is_same_address(state.contract_address, channel_manager_address):
        click.echo('Channel manager contract address mismatch')
        sys.exit(1)

    click.echo('Closing all open channels with valid balance proofs for '
               'receiver {}'.format(receiver_address))
    close_open_channels(private_key, state, channel_manager_contract)


def close_open_channels(
        private_key: str,
        state: ChannelManagerState,
        channel_manager_contract: Contract,
        repetitions=None,
        wait=lambda: gevent.sleep(1)
):
    web3 = channel_manager_contract.web3
    channels_with_balance_proof = [
        c for c in state.channels.values() if c.last_signature is not None
    ]
    n_channels = len(state.channels)
    n_no_balance_proof = len(state.channels) - len(channels_with_balance_proof)
    n_txs_sent = 0
    pending_txs = {}

    if repetitions:
        iterator = range(repetitions)
    else:
        iterator = count()
    for _ in iterator:
        n_non_existant = 0
        n_invalid_balance_proof = 0
        for channel in channels_with_balance_proof:
            # lookup channel on block chain
            channel_id = (channel.sender, channel.receiver, channel.open_block_number)
            try:
                channel_info = channel_manager_contract.call().getChannelInfo(*channel_id)
            except (BadFunctionCallOutput, TransactionFailed):
                n_non_existant += 1
                continue
            _, deposit, settle_block_number, closing_balance = channel_info

            is_valid = channel.balance <= deposit
            n_invalid_balance_proof += int(not is_valid)
            close_sent = (channel.sender, channel.open_block_number) in pending_txs

            # send close if open or settling with wrong balance, unless already done
            if not close_sent and is_valid:
                raw_tx = create_signed_contract_transaction(
                    private_key,
                    channel_manager_contract,
                    'uncooperativeClose',
                    [
                        channel.receiver,
                        channel.open_block_number,
                        channel.balance,
                        decode_hex(channel.last_signature)
                    ]
                )
                tx_hash = web3.eth.sendRawTransaction(raw_tx)
                log.info('sending close tx (hash: {})'.format(tx_hash))
                pending_txs[channel.sender, channel.open_block_number] = tx_hash
                n_txs_sent += 1

        # print status
        msg_status = 'block: {}, pending txs: {}, total txs sent: {}'
        msg_progress = (
            'initial channels: {}, settled: {}, pending txs: {}, no BP: {}, invalid BP: {}'
        )
        log.info(msg_status.format(web3.eth.blockNumber, len(pending_txs), n_txs_sent))
        log.info(msg_progress.format(
            n_channels,
            n_non_existant,
            len(pending_txs),
            n_no_balance_proof,
            n_invalid_balance_proof
        ))

        # wait for next block
        block_before = web3.eth.blockNumber
        while web3.eth.blockNumber == block_before:
            wait()

        # update pending txs
        confirmed = []
        for channel_id, tx_hash in pending_txs.items():
            receipt = web3.eth.getTransactionReceipt(tx_hash)
            if receipt is None:
                continue
            tx = web3.eth.getTransaction(tx_hash)
            if receipt.gasUsed == tx.gas:
                raise ValueError('Transaction failed, out of gas (hash: {})'.format(tx_hash))
            confirmed.append(channel_id)
        for channel_id in confirmed:
            pending_txs.pop(channel_id)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    main()
